#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)


import time

from fabrictestbed.slice_manager import Status, SliceState
from .abc_utils import AbcUtils


class SliceUtils(AbcUtils):
    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

    @staticmethod
    def delete_all_with_substring(string):
        slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))

        slices = list(filter(lambda x: string in x.slice_name, slices))

        for slice in slices:
            return_status, result = slice_manager.delete(slice_object=slice)
            print(
                "Deleting Slice: {}.  Response Status {}".format(
                    slice.slice_name, return_status
                )
            )

    @staticmethod
    def delete_all():
        slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))

        for slice in slices:
            return_status, result = slice_manager.delete(slice_object=slice)
            print(
                "Deleting Slice: {}.  Response Status {}".format(
                    slice.slice_name, return_status
                )
            )

    @staticmethod
    def delete_slice(slice_name=None, slice_id=None):
        slice = SliceUtils.get_slice(slice_name=slice_name, slice_id=slice_id)

        slice_manager = AbcUtils.create_slice_manager()
        return_status, result = slice_manager.delete(slice_object=slice)
        print(
            "Deleting Slice: {}.  Response Status {}".format(
                slice.slice_name, return_status
            )
        )

    @staticmethod
    def get_slice(slice_name=None, slice_id=None, slice_manager=None):
        if not slice_manager:
            slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))
        try:
            if slice_id:
                slice = list(filter(lambda x: x.slice_id == slice_id, slices))[0]
            elif slice_name:
                slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
            else:
                raise Exception(
                    "Slice not found. Slice name or id requried. name: {}, slice_id: {}".format(
                        str(slice_name), str(slice_id)
                    )
                )
        except:
            raise Exception(
                "Slice not found name: {}, slice_id: {}".format(
                    str(slice_name), str(slice_id)
                )
            )

        return slice

    @staticmethod
    def list_all_slices(excludes=[SliceState.Dead, SliceState.Closing]):
        slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))
        try:
            for slice in slices:
                print(
                    "{} | {} | {}".format(
                        slice.slice_name, slice.slice_id, slice.slice_state
                    )
                )
        except:
            raise Exception(
                "Slice not found name: {}, slice_id: {}".format(
                    str(slice_name), str(slice_id)
                )
            )

        return slice

    @staticmethod
    def get_slice_error(slice_id):
        slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(
            includes=[SliceState.Dead, SliceState.Closing, SliceState.StableError]
        )
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))
        try:
            if slice_id:
                slice = list(filter(lambda x: x.slice_id == slice_id, slices))[0]
            else:
                raise Exception(
                    "Slice not found. Slice name or id requried. slice_id: {}".format(
                        str(slice_id)
                    )
                )
        except Exception as e:
            print("Exception: {}".format(str(e)))
            raise Exception("Slice not found slice_id: {}".format(str(slice_id)))

        return_status, slivers = slice_manager.slivers(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("Failed to get slivers: {}".format(slivers))

        return_errors = []
        for s in slivers:
            status, sliver_status = slice_manager.sliver_status(sliver=s)

            # print("Response Status {}".format(status))
            if status == Status.OK:
                # print()
                # print("Sliver Status {}".format(sliver_status))
                # print()
                # return_errors.append("Sliver: {} {}".format(s.name))
                # return_errors.append(sliver_status.notices)
                return_errors.append(sliver_status)

        return return_errors

    @staticmethod
    def wait_for_slice(
        slice_name=None, slice_id=None, timeout=360, interval=10, progress=False
    ):
        slice_manager = AbcUtils.create_slice_manager()
        slice = SliceUtils.get_slice(slice_name=slice_name, slice_id=slice_id)

        timeout_start = time.time()

        if progress:
            print("Waiting for slice .", end="")
        while time.time() < timeout_start + timeout:
            return_status, slices = slice_manager.slices(
                excludes=[SliceState.Dead, SliceState.Closing]
            )

            if return_status == Status.OK:
                slice = list(
                    filter(lambda x: x.slice_name == slice.slice_name, slices)
                )[0]
                if slice.slice_state == "StableOK":
                    if progress:
                        print(" Slice state: {}".format(slice.slice_state))
                    return slice
                if (
                    slice.slice_state == "Closing"
                    or slice.slice_state == "Dead"
                    or slice.slice_state == "StableError"
                ):
                    if progress:
                        print(" Slice state: {}".format(slice.slice_state))
                    return slice
            else:
                print(f"Failure: {slices}")

            if progress:
                print(".", end="")
            time.sleep(interval)

        if time.time() >= timeout_start + timeout:
            if progress:
                print(
                    " Timeout exceeded ({} sec). Slice: {} ({})".format(
                        timeout, slice.slice_name, slice.slice_state
                    )
                )
            return slice



# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import os

from abc import ABC
from fabrictestbed.util.constants import Constants


class AbcFabLIB(ABC):
    FABRIC_BASTION_USERNAME = "FABRIC_BASTION_USERNAME"
    FABRIC_BASTION_KEY_LOCATION = "FABRIC_BASTION_KEY_LOCATION"
    FABRIC_BASTION_HOST = "FABRIC_BASTION_HOST"
    FABRIC_BASTION_KEY_PASSWORD = "FABRIC_BASTION_KEY_PASSWORD"
    FABRIC_BASTION_HOST_PRIVATE_IPV4 = "FABRIC_BASTION_HOST_PRIVATE_IPV4"
    FABRIC_BASTION_HOST_PRIVATE_IPV6 = "FABRIC_BASTION_HOST_PRIVATE_IPV6"
    FABRIC_SLICE_PUBLIC_KEY_FILE = "FABRIC_SLICE_PUBLIC_KEY_FILE"
    FABRIC_SLICE_PRIVATE_KEY_FILE = "FABRIC_SLICE_PRIVATE_KEY_FILE"
    FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE = "FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE"

    def __init__(self):
        """
        Constructor. Sets environment variables for important FABRIC values.
        """
        self.credmgr_host = None
        self.orchestrator_host = None
        self.fabric_token = None
        self.project_id = None

        self.bastion_username = None
        self.bastion_key_filename = None
        self.bastion_public_addr = None
        self.bastion_private_ipv4_addr = "0.0.0.0"
        self.bastion_private_ipv6_addr = "0:0:0:0:0:0"

        self.slice_keys = {}
        self.default_slice_key = {}
        self.slice_keys["default"] = self.default_slice_key

        if Constants.FABRIC_CREDMGR_HOST in os.environ:
            self.credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]

        if Constants.FABRIC_ORCHESTRATOR_HOST in os.environ:
            self.orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]

        if Constants.FABRIC_TOKEN_LOCATION in os.environ:
            self.fabric_token = os.environ[Constants.FABRIC_TOKEN_LOCATION]

        if Constants.FABRIC_PROJECT_ID in os.environ:
            self.project_id = os.environ[Constants.FABRIC_PROJECT_ID]

        # Basstion host setup
        if self.FABRIC_BASTION_USERNAME in os.environ:
            self.bastion_username = os.environ[self.FABRIC_BASTION_USERNAME]
        if self.FABRIC_BASTION_KEY_LOCATION in os.environ:
            self.bastion_key_filename = os.environ[self.FABRIC_BASTION_KEY_LOCATION]
        if self.FABRIC_BASTION_HOST in os.environ:
            self.bastion_public_addr = os.environ[self.FABRIC_BASTION_HOST]
        if self.FABRIC_BASTION_HOST_PRIVATE_IPV4 in os.environ:
            self.bastion_private_ipv4_addr = os.environ[
                self.FABRIC_BASTION_HOST_PRIVATE_IPV4
            ]
        if self.FABRIC_BASTION_HOST_PRIVATE_IPV6 in os.environ:
            self.bastion_private_ipv6_addr = os.environ[
                self.FABRIC_BASTION_HOST_PRIVATE_IPV6
            ]

        # Slice Keys
        if self.FABRIC_SLICE_PUBLIC_KEY_FILE in os.environ:
            self.default_slice_key["slice_public_key_file"] = os.environ[
                self.FABRIC_SLICE_PUBLIC_KEY_FILE
            ]
            with open(os.environ[self.FABRIC_SLICE_PUBLIC_KEY_FILE], "r") as fd:
                self.default_slice_key["slice_public_key"] = fd.read().strip()
        if self.FABRIC_SLICE_PRIVATE_KEY_FILE in os.environ:
            # self.slice_private_key_file=os.environ['FABRIC_SLICE_PRIVATE_KEY_FILE']
            self.default_slice_key["slice_private_key_file"] = os.environ[
                self.FABRIC_SLICE_PRIVATE_KEY_FILE
            ]
        if "FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE" in os.environ:
            # self.slice_private_key_passphrase = os.environ['FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE']
            self.default_slice_key["slice_private_key_passphrase"] = os.environ[
                self.FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE
            ]



import os
import pathlib
import tempfile
import unittest

from fabrictestbed.util.constants import Constants

from fabrictestbed_extensions.fablib.fablib import FablibManager


class FablibManagerTests(unittest.TestCase):
    """
    Some basic tests to exercise FablibManager.

    Instantiating FablibManager will throw some exceptions if some
    required environment variables are not set.
    """

    required_env_vars = [
        Constants.FABRIC_CREDMGR_HOST,
        Constants.FABRIC_ORCHESTRATOR_HOST,
        Constants.FABRIC_PROJECT_ID,
        # Constants.FABRIC_TOKEN_LOCATION,
        FablibManager.FABRIC_BASTION_HOST,
        FablibManager.FABRIC_BASTION_USERNAME,
        FablibManager.FABRIC_BASTION_KEY_LOCATION,
    ]

    def setUp(self):
        # Run each test with an empty environment.
        os.environ.clear()

    def test_fablib_manager_no_env_vars(self):
        # Test with no required env vars set.
        with self.assertRaises(AttributeError):
            FablibManager()

    def test_fablib_manager_one_env_var(self):
        # Test with some required env vars set.
        for var in self.required_env_vars:
            with self.assertRaises(AttributeError):
                os.environ[var] = "dummy"
                FablibManager()

    def test_fablib_manager_all_env_vars(self):
        # Test with all required configuration except
        # FABRIC_TOKEN_LOCATION.
        for var in self.required_env_vars:
            os.environ[var] = "dummy"

        with self.assertRaises(AttributeError):
            FablibManager()

    def test_fablib_manager_test_only_cm_host(self):
        with self.assertRaises(AttributeError):
            os.environ[Constants.FABRIC_CREDMGR_HOST] = "dummy"
            FablibManager()

    def test_fablib_manager_test_only_orchestrator_host(self):
        with self.assertRaises(AttributeError):
            os.environ[Constants.FABRIC_ORCHESTRATOR_HOST] = "dummy"
            FablibManager()

    def test_fablib_manager_test_only_project_id(self):
        with self.assertRaises(AttributeError):
            os.environ[Constants.FABRIC_PROJECT_ID] = "dummy"
            FablibManager()

    def test_fablib_manager_test_only_token_location(self):
        with self.assertRaises(AttributeError):
            os.environ[Constants.FABRIC_TOKEN_LOCATION] = "dummy"
            FablibManager()

    def test_fablib_manager_test_with_no_token_file(self):
        # Should fail when token location is not a valid path.

        # set all required env vars.
        for var in self.required_env_vars:
            os.environ[var] = "dummy"

        os.environ[Constants.FABRIC_TOKEN_LOCATION] = "dummy"

        # FablibManager() without a valid token or token location
        # should raise a "ValueError: Invalid value for
        # `refresh_token`, must not be `None`"
        with self.assertRaises(ValueError):
            FablibManager()

    def test_fablib_manager_test_with_dummy_token(self):
        # TODO: That FablibManager() calls build_slice_manager()
        # complicates writing a test for it.  It eventually makes a
        # network call to credential manager API, but it is not right
        # for a unit test to do such a thing.  We could probably
        # somehow mock a CM here?

        # set all required env vars.
        for var in self.required_env_vars:
            os.environ[var] = "dummy"

        # '.invalid' is an invalid host per RFC 6761, so this test
        # must fail without ever making a successful network call.
        os.environ[Constants.FABRIC_CREDMGR_HOST] = ".invalid"
        path = os.path.join(os.path.dirname(__file__), "dummy-token.json")
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = path

        with self.assertRaises(Exception):
            FablibManager()

    def test_fablib_manager_with_empty_config(self):
        # Check that an empty configuration file will cause
        # FablibManager to raise an error.
        rcfile = tempfile.NamedTemporaryFile()
        rcfile.flush()

        with self.assertRaises(AttributeError):
            FablibManager(fabric_rc=rcfile.name)

    def test_fablib_manager_with_some_config(self):
        # Test with some configuration in the rc file.
        rcfile = tempfile.NamedTemporaryFile()

        # Write all configuration except FABRIC_TOKEN_LOCATION.  The
        # token location is a special case because when it is set,
        # Fablib will need network access to reach CredentialManager.
        for var in self.required_env_vars:
            rcfile.write(f"export {var} = dummy\n".encode())

        rcfile.flush()

        with self.assertRaises(AttributeError) as ctx:
            FablibManager(fabric_rc=rcfile.name)

        # Check that the error is what we expected.
        self.assertIsInstance(ctx.exception, AttributeError)

        # Check that the error message is what we expected: the only
        # error should be about missing token.
        self.assertEqual(
            str(ctx.exception),
            "Error initializing FablibManager: ['FABRIC token is not set']",
        )



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import (
    Labels,
    ExperimentTopology,
    Capacities,
    ComponentType,
    ComponentModelType,
    ServiceType,
    ComponentCatalog,
)
from fabrictestbed.slice_editor import ExperimentTopology, Capacities
from fabrictestbed.slice_manager import SliceManager, Status, SliceState


from .abc_test import AbcTest

from .. import images


class ComponentTest(AbcTest):
    COMPONENT_MODELS = {
        "GPU_Tesla_T4": ComponentModelType.GPU_Tesla_T4,
        "GPU_RTX6000": ComponentModelType.GPU_RTX6000,
        "GPU_A30": ComponentModelType.GPU_A30,
        "GPU_A40": ComponentModelType.GPU_A40,
        "SharedNIC_ConnectX_6": ComponentModelType.SharedNIC_ConnectX_6,
        "SmartNIC_ConnectX_6": ComponentModelType.SmartNIC_ConnectX_6,
        "SmartNIC_ConnectX_5": ComponentModelType.SmartNIC_ConnectX_5,
        "NVME_P4510": ComponentModelType.NVME_P4510,
        "FPGA_Xilinx_U280": ComponentModelType.FPGA_Xilinx_U280
    }

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

    def create_slice(
        self, name_prefix="component_test", model_type=None, site_name="", node_count=1
    ):
        image = "default_centos_8"
        image_type = "qcow2"
        cores = 2
        ram = 2
        disk = 2
        # model_type = ComponentModelType.GPU_Tesla_T4
        # Tesla_T4_component_type = ComponentType.Tesla_T4
        # Tesla_T4_model = 'P4510'
        # Tesla_T4_name = 'gpu1'
        component_name = "dev1"
        component_model_name = str(model_type).split(".", 1)[1]

        topology = ExperimentTopology()

        cap = Capacities(core=cores, ram=ram, disk=disk)

        slice_name = name_prefix + "-" + component_model_name + "-" + str(site_name)

        for node_num in range(0, node_count):
            # node_name=slice_name+"-"+str(node_num)
            node_name = "node" + str(node_num)
            # Add node
            n1 = topology.add_node(name=node_name, site=site_name)

            # Set Properties
            n1.set_properties(capacities=cap, image_type=image_type, image_ref=image)

            # Add the PCI Tesla_T4 device
            n1.add_component(model_type=model_type, name=component_name)
            # n1.add_component(ctype=Tesla_T4_component_type, model=Tesla_T4_model, name=Tesla_T4_name)

        # Generate Slice Graph
        slice_graph = topology.serialize()

        # Request slice from Orchestrator
        return_status, slice_reservations = self.slice_manager.create(
            slice_name=slice_name, slice_graph=slice_graph, ssh_key=self.node_ssh_key
        )

        if return_status == Status.OK:
            slice_id = slice_reservations[0].get_slice_id()
            # print("Submitted slice creation request. Slice ID: {}".format(slice_id))
        else:
            print(f"Failure: {slice_reservations}")

        # time.sleep(30)

        return_status, slices = self.slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )

        if return_status == Status.OK:
            self.slices.append(
                list(filter(lambda x: x.slice_name == slice_name, slices))[0]
            )
            # self.slice = self.wait_for_slice(slice, progress=True)

        # print()
        # print("Slice Name : {}".format(self.slice.slice_name))
        # print("ID         : {}".format(self.slice.slice_id))
        # print("State      : {}".format(self.slice.slice_state))
        # print("Lease End  : {}".format(self.slice.lease_end))

    def config_test_GPU_RTX6000(self, node):
        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_GPU_RTX6000(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")
            self.config_test_GPU_RTX6000(node)
            try:
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                expected_stdout = "3D controller: NVIDIA Corporation TU102GL [Quadro RTX 6000/8000] (rev a1)"

                script = "#!/bin/bash  \n" """lspci | grep NVIDIA \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "... | ", end="")
                if expected_stdout in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_NVME_P4510(self, node):
        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_NVME_P4510(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")
            self.config_test_NVME_P4510(node)
            try:
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                # There are two types of NVMe
                expected_stdout1 = "Non-Volatile memory controller: Intel Corporation NVMe Datacenter SSD [3DNAND, Beta Rock Controller]"
                expected_stdout2 = "Non-Volatile memory controller: Toshiba Corporation NVMe SSD Controller Cx5 (rev 01)"
                script = "#!/bin/bash  \n" """lspci | grep NVMe \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "... | ", end="")
                if expected_stdout1 in stdout_str or expected_stdout2 in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_SmartNIC_ConnectX_5(self, node):
        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_SmartNIC_ConnectX_5(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")
            self.config_test_SmartNIC_ConnectX_5(node)
            try:
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                expected_stdout = "Ethernet controller: Mellanox Technologies MT27800 Family [ConnectX-5]"

                script = "#!/bin/bash  \n" """lspci | grep ConnectX \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "... | ", end="")
                if expected_stdout in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_SmartNIC_ConnectX_6(self, node):
        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_SmartNIC_ConnectX_6(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            # print(" Test: SmartNIC_ConnectX_6 | " + str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")

            self.config_test_SmartNIC_ConnectX_6(node)
            try:
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                expected_stdout = "Ethernet controller: Mellanox Technologies MT28908 Family [ConnectX-6]"

                script = "#!/bin/bash  \n" """lspci | grep ConnectX-6 \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "...  | ", end="")
                if expected_stdout in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_SharedNIC_ConnectX_6(self, node):
        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_SharedNIC_ConnectX_6(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")
            self.config_test_SharedNIC_ConnectX_6(node)
            try:
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                expected_stdout = "Ethernet controller: Mellanox Technologies MT28908 Family [ConnectX-6 Virtual Function]"

                script = "#!/bin/bash  \n" """lspci | grep ConnectX-6 \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "... | ", end="")
                if expected_stdout in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_GPU_Tesla_T4(self, node):
        node_name = node.name
        # print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
        # print("Node:")
        # print("Config | ", end = '')
        # print(str(node_name) + " | ", end = '')
        try:
            # print("{}".format(str(node_name)) + " | ", end = '')

            # print("   Name              : {}".format(node_name))
            # print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
            # print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
            # print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
            # print("   Image             : {}".format(node.image_ref))
            # print("   Image Type        : {}".format(node.image_type))
            # print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
            # print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
            # print("   Site              : {}".format(node.site))
            # print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

            # print("   Management IP     : {}".format(node.management_ip))
            # print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
            # print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
            # print("   Components        : {}".format(node.components))
            # print("   Interfaces        : {}".format(node.interfaces))
            # print()
            name = node_name
            management_ip = node.management_ip

            # print("------------------------- Test Output ---------------------------")
            script = "#!/bin/bash  \n" "sudo yum install -y pciutils \n"
            try:
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print("Config Done | ", end="")
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end="")
            # print(str(stdout_str.replace('\n','')) + " | ", end = '')
            # print("-----------------------------------------------------------------")
            # stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
        except Exception as e:
            print("Error in test: Error {}".format(e), end="")
            traceback.print_exc()

    def run_test_GPU_Tesla_T4(self, slice):
        # self.config_test(slice)

        return_status, topology = self.slice_manager.get_slice_topology(
            slice_object=slice
        )
        if return_status != Status.OK:
            raise Exception(
                "config_test failed to get topology. slice; {}, error {}".format(
                    str(slice), str(topology)
                )
            )

        for node_name, node in topology.nodes.items():
            # print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            # print("Node:")
            # print(str(node_name) + " | ", end = '')
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end="")
            self.config_test_GPU_Tesla_T4(node)
            try:
                # print("   Name              : {}".format(node_name))
                # print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
                # print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
                # print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
                # print("   Image             : {}".format(node.image_ref))
                # print("   Image Type        : {}".format(node.image_type))
                # print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
                # print("   Site              : {}".format(node.site))
                # print("   Management IP     : {}".format(node.management_ip))
                # print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
                # print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
                # print("   Components        : {}".format(node.components))
                # print("   Interfaces        : {}".format(node.interfaces))
                # print()
                print(
                    "{}".format(
                        str(
                            node.get_property(pname="label_allocations").instance_parent
                        )
                    ).replace("\n", "")
                    + " | ",
                    end="",
                )
                print(
                    "{}".format(str(node.management_ip)).replace("\n", "") + " | ",
                    end="",
                )

                name = node_name
                reservation_id = node.get_property(
                    pname="reservation_info"
                ).reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + "-" + name.lower()

                expected_stdout = (
                    "3D controller: NVIDIA Corporation TU104GL [Tesla T4] (rev a1)"
                )

                # print("------------------------- Test Output ---------------------------")
                script = "#!/bin/bash  \n" """lspci | grep Tesla \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(
                    node_username="centos", node=node, script=script
                )
                print(str(stdout_str.replace("\n", ""))[0:45] + "... | ", end="")
                # print("-----------------------------------------------------------------")
                # stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
                if expected_stdout in stdout_str:
                    print("Success")
                else:
                    print("Fail")
                    print(stdout_str)
            except Exception as e:
                print("Error in test: Error {}".format(e))
                traceback.print_exc()

    def delete_slice(self):
        status, result = self.slice_manager.delete(slice_id=self.slice.slice_id)

        # print("Response Status {}".format(status))
        # print("Response received {}".format(result))

    def get_existing_slice(self, slice_name):
        # Get existing slice
        return_status, slices = self.slice_manager.slices(
            excludes=[SliceState.Dead, SliceState.Closing]
        )

        if return_status == Status.OK:
            # for slice in slices:
            #    print("{}:".format(slice.slice_name))
            #    print("   ID         : {}".format(slice.slice_id))
            #    print("   State      : {}".format(slice.slice_state))
            #    print("   Lease End  : {}".format(slice.lease_end))
            # print("got Slice")
            pass
        else:
            print(f"Failure: {slices}")

        self.slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
        status, self.topology = self.slice_manager.get_slice_topology(
            slice_object=self.slice
        )
        if return_status == Status.OK:
            # print("got topology")
            pass
        else:
            print(f"Failure: {self.topology}")

        # print("Slice Name : {}".format(self.slice.slice_name))
        # print("ID         : {}".format(self.slice.slice_id))
        # print("State      : {}".format(self.slice.slice_state))
        # print("Lease End  : {}".format(self.slice.lease_end))

    def run_all(
        self,
        test_name="component_test",
        tests=[],
        create_slice=True,
        run_test=True,
        delete=True,
        node_count=1,
    ):
        self.test_name = test_name

        # Create slices
        if create_slice:
            for site, model_type, node_count in tests:
                print(
                    "Create slice:  site: {}, component: {}, node_count: {}. Nodes: ".format(
                        site, model_type, node_count
                    ),
                    end="",
                )
                # print("Creating Slice")
                for node_num in range(0, node_count):
                    try:
                        print(" {}".format(node_num), end="")

                        self.create_slice(
                            name_prefix=test_name + "-" + str(node_num),
                            model_type=ComponentTest.COMPONENT_MODELS[model_type],
                            site_name=site,
                            node_count=1,
                        )
                        # time.sleep(2)

                        # self.slice = self.wait_for_slice(self.slice, progress=True, timeout=300+(20*node_count))
                        # time.sleep(10)

                    except Exception as e:
                        print("Create Slice FAILED. Error {}".format(e))
                        traceback.print_exc()
                print(", Done")

            for slice in self.slices:
                try:
                    s = self.wait_for_slice(slice, progress=True, timeout=300)
                except Exception as e:
                    print("Slice failed wait_for_slice. {}".format(slice))
                    print(str(e))
                    # self.slices.remove(slice)

            time.sleep(120)
            # test ssh conectivity
            for slice in self.slices:
                try:
                    self.run_ssh_test(slice)
                except Exception as e:
                    # print("Failed to get topology: {}".format(slice))
                    # print(str(e))
                    print(
                        " Test: ssh | "
                        + str(slice.slice_name)
                        + " | slice closing | Fail "
                    )

        # Run tests
        if run_test:
            print("Run tests")
            for site, model_type, node_count in tests:
                # print("Run test:  site: {}, component: {}, node_count: {}".format(site,model_type,node_count))
                # print("Creating Slice")
                for node_num in range(0, node_count):
                    try:
                        # print("Running test  site: {}, component: {}, node_num: {}".format(site,model_type,node_num))
                        name_prefix = test_name + "-" + str(node_num)
                        component_model_name = str(model_type)  # .split(".",1)[1]
                        slice_name = (
                            name_prefix + "-" + component_model_name + "-" + str(site)
                        )
                        # print("Running test, slice_name: {}".format(slice_name))
                        from fabrictestbed_extensions.utils.slice import SliceUtils

                        slice = SliceUtils.get_slice(
                            slice_name=slice_name, slice_manager=self.slice_manager
                        )

                        if model_type == "GPU_Tesla_T4":
                            self.run_test_GPU_Tesla_T4(slice)
                        elif model_type == "GPU_RTX6000":
                            self.run_test_GPU_RTX6000(slice)
                        elif model_type == "SharedNIC_ConnectX_6":
                            self.run_test_SharedNIC_ConnectX_6(slice)
                        elif model_type == "SmartNIC_ConnectX_6":
                            self.run_test_SmartNIC_ConnectX_6(slice)
                        elif model_type == "SmartNIC_ConnectX_5":
                            self.run_test_SmartNIC_ConnectX_5(slice)
                        elif model_type == "NVME_P4510":
                            self.run_test_NVME_P4510(slice)
                        else:
                            print(
                                " Test: XXX | "
                                + str(slice_name)
                                + " | unknown test | Skipped"
                            )

                    except Exception as e:
                        print(
                            " Test: XXX | "
                            + str(slice_name)
                            + " | "
                            + str(e)
                            + " | Failed"
                        )

                        # print("Run test FAILED. Error {}".format(e))
                        # traceback.print_exc()

        # self.run(slice_name + "-" + site, create_slice=False, run_test=run_test, delete=delete, site=site, node_count=node_count)

    def run(
        self,
        test_name="component_test",
        create_slice=True,
        run_test=True,
        delete=True,
        site=None,
        node_count=1,
    ):
        """
        Run the test
        :return:
        """
        self.test_name = test_name
        print("Tesla_T4 test, slice_name: {}, site: {}".format(slice_name, site))
        if create_slice:
            # print("Creating Slice")
            try:
                self.create_slice(
                    model_type=ComponentModelType.GPU_Tesla_T4,
                    site_name=site,
                    node_count=node_count,
                )
                # time.sleep(5)

                self.slice = self.wait_for_slice(
                    self.slice, progress=True, timeout=300 + (20 * node_count)
                )
                # time.sleep(10)

            except Exception as e:
                print("Create Slice FAILED. Error {}".format(e))
                traceback.print_exc()
            time.sleep(5 * node_count)

        self.get_existing_slice(slice_name)
        # time.sleep(60)

        if run_test:
            try:
                # print("Run Test")
                self.run_test_tesla_t4()
            except Exception as e:
                print("Run test FAILED. Error {}".format(e))
                traceback.print_exc()
        if delete:
            try:
                # print("Delete Slice")
                self.delete_slice()
            except:
                print("Delete FAILED")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
from __future__ import annotations
import logging
from tabulate import tabulate

from typing import List, Tuple
import json

from fabrictestbed.slice_editor import AdvertisedTopology
from fabrictestbed.slice_editor import Capacities
from fabrictestbed.slice_manager import Status
from fim.user import link, interface


class Resources:
    site_pretty_names = {
        "name": "Name",
        "state": "State",
        "address": "Address",
        "location": "Location",
        "hosts": "Hosts",
        "cpus": "CPUs",
        "cores_available": "Cores Available",
        "cores_capacity": "Cores Capacity",
        "cores_allocated": "Cores Allocated",
        "ram_available": "RAM Available",
        "ram_capacity": "RAM Capacity",
        "ram_allocated": "RAM Allocated",
        "disk_available": "Disk Available",
        "disk_capacity": "Disk Capacity",
        "disk_allocated": "Disk Allocated",
        "nic_basic_available": "Basic NIC Available",
        "nic_basic_capacity": "Basic NIC Capacity",
        "nic_basic_allocated": "Basic NIC Allocated",
        "nic_connectx_6_available": "ConnectX-6 Available",
        "nic_connectx_6_capacity": "ConnectX-6 Capacity",
        "nic_connectx_6_allocated": "ConnectX-6 Allocated",
        "nic_connectx_5_available": "ConnectX-5 Available",
        "nic_connectx_5_capacity": "ConnectX-5 Capacity",
        "nic_connectx_5_allocated": "ConnectX-5 Allocated",
        "nvme_available": "NVMe Available",
        "nvme_capacity": "NVMe Capacity",
        "nvme_allocated": "NVMe Allocated",
        "tesla_t4_available": "Tesla T4 Available",
        "tesla_t4_capacity": "Tesla T4 Capacity",
        "tesla_t4_allocated": "Tesla T4 Allocated",
        "rtx6000_available": "RTX6000 Available",
        "rtx6000_capacity": "RTX6000 Capacity",
        "rtx6000_allocated": "RTX6000 Allocated",
        "a30_available": "A30 Available",
        "a30_capacity": "A30 Capacity",
        "a30_allocated": "A30 Allocated",
        "a40_available": "A40 Available",
        "a40_capacity": "A40 Capacity",
        "a40_allocated": "A40 Allocated",
    }

    def __init__(self, fablib_manager, force_refresh: bool = False):
        """
        Constructor
        :return:
        """
        super().__init__()

        self.fablib_manager = fablib_manager

        self.topology = None

        self.update(force_refresh=force_refresh)

    def __str__(self) -> str:
        """
        Creates a tabulated string of all the available resources.

        Intended for printing available resources.

        :return: Tabulated string of available resources
        :rtype: String
        """
        table = []
        for site_name, site in self.topology.sites.items():
            # logging.debug(f"site -- {site}")
            table.append(
                [
                    site.name,
                    self.get_cpu_capacity(site_name),
                    f"{self.get_core_available(site_name)}/{self.get_core_capacity(site_name)}",
                    f"{self.get_ram_available(site_name)}/{self.get_ram_capacity(site_name)}",
                    f"{self.get_disk_available(site_name)}/{self.get_disk_capacity(site_name)}",
                    # self.get_host_capacity(site_name),
                    # self.get_location_postal(site_name),
                    # self.get_location_lat_long(site_name),
                    f"{self.get_component_available(site_name,'SharedNIC-ConnectX-6')}/{self.get_component_capacity(site_name,'SharedNIC-ConnectX-6')}",
                    f"{self.get_component_available(site_name,'SmartNIC-ConnectX-6')}/{self.get_component_capacity(site_name,'SmartNIC-ConnectX-6')}",
                    f"{self.get_component_available(site_name,'SmartNIC-ConnectX-5')}/{self.get_component_capacity(site_name,'SmartNIC-ConnectX-5')}",
                    f"{self.get_component_available(site_name,'NVME-P4510')}/{self.get_component_capacity(site_name,'NVME-P4510')}",
                    f"{self.get_component_available(site_name,'GPU-Tesla T4')}/{self.get_component_capacity(site_name,'GPU-Tesla T4')}",
                    f"{self.get_component_available(site_name,'GPU-RTX6000')}/{self.get_component_capacity(site_name,'GPU-RTX6000')}",
                    f"{self.get_component_available(site_name, 'GPU-A30')}/{self.get_component_capacity(site_name, 'GPU-A30')}",
                    f"{self.get_component_available(site_name, 'GPU-A40')}/{self.get_component_capacity(site_name, 'GPU-A40')}",
                    f"{self.get_component_available(site_name, 'FPGA-Xilinx-U280')}/{self.get_component_capacity(site_name, 'FPGA-Xilinx-U280')}",
                ]
            )

        return tabulate(
            table,
            headers=[
                "Name",
                "CPUs",
                "Cores",
                f"RAM ({Capacities.UNITS['ram']})",
                f"Disk ({Capacities.UNITS['disk']})",
                # "Workers"
                # "Physical Address",
                # "Location Coordinates"
                "Basic (100 Gbps NIC)",
                "ConnectX-6 (100 Gbps x2 NIC)",
                "ConnectX-5 (25 Gbps x2 NIC)",
                "P4510 (NVMe 1TB)",
                "Tesla T4 (GPU)",
                "RTX6000 (GPU)",
                "A30 (GPU)",
                "A40 (GPU)",
                "FPGA-Xilinx-U280"
            ],
        )

    def show_site(
        self,
        site_name: str,
        output: str = None,
        fields: list[str] = None,
        quiet: bool = False,
        pretty_names=True,
    ) -> str:
        """
        Creates a tabulated string of all the available resources at a specific site.

        Intended for printing available resources at a site.

        :param site_name: site name
        :type site_name: String
        :return: Tabulated string of available resources
        :rtype: String
        """
        site = self.topology.sites[site_name]

        data = self.site_to_dict(site)

        if pretty_names:
            pretty_names_dict = self.site_pretty_names
        else:
            pretty_names_dict = {}

        site_table = self.get_fablib_manager().show_table(
            data,
            fields=fields,
            title="Site",
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return site_table

    def get_site_names(self) -> List[str]:
        """
        Gets a list of all currently available site names

        :return: list of site names
        :rtype: List[String]
        """
        site_name_list = []
        for site_name in self.topology.sites.keys():
            site_name_list.append(str(site_name))

        return site_name_list

    def get_topology_site(self, site_name: str) -> str:
        """
        Not recommended for most users.
        """
        try:
            return self.topology.sites[site_name]
        except Exception as e:
            logging.warning(f"Failed to get site {site_name}")
            return ""

    def get_state(self, site_name: str):
        try:
            return str(
                self.get_topology_site(site_name)
                .get_property("maintenance_info")
                .get(site_name)
                .state
            )
        except Exception as e:
            logging.warning(f"Failed to get site state {site_name}")
            return ""

    def get_component_capacity(self, site_name: str, component_model_name: str) -> int:
        """
        Gets gets the total site capacity of a component by model name.

        :param site_name: site name
        :type site_name: String
        :param component_model_name: component model name
        :type component_model_name: String
        :return: total component capacity
        :rtype: int
        """
        try:
            return (
                self.get_topology_site(site_name)
                .components[component_model_name]
                .capacities.unit
            )
        except Exception as e:
            # logging.debug(f"Failed to get {component_model_name} capacity {site_name}")
            return 0

    def get_component_allocated(self, site_name: str, component_model_name: str) -> int:
        """
        Gets gets number of currently allocated components on a the site
        by the component by model name.

        :param site_name: site name
        :type site_name: String
        :param component_model_name: component model name
        :type component_model_name: String
        :return: currently allocated component of this model
        :rtype: int
        """
        try:
            return (
                self.get_topology_site(site_name)
                .components[component_model_name]
                .capacity_allocations.unit
            )
        except Exception as e:
            # logging.debug(f"Failed to get {component_model_name} allocated {site_name}")
            return 0

    def get_component_available(self, site_name: str, component_model_name: str) -> int:
        """
        Gets gets number of currently available components on the site
        by the component by model name.

        :param site_name: site name
        :type site_name: String
        :param component_model_name: component model name
        :type component_model_name: String
        :return: currently available component of this model
        :rtype: int
        """
        try:
            return self.get_component_capacity(
                site_name, component_model_name
            ) - self.get_component_allocated(site_name, component_model_name)
        except Exception as e:
            # logging.debug(f"Failed to get {component_model_name} available {site_name}")
            return self.get_component_capacity(site_name, component_model_name)

    def get_location_lat_long(self, site_name: str) -> Tuple[float, float]:
        """
        Gets gets location of a site in latitude and longitude

        :param site_name: site name
        :type site_name: String
        :return: latitude and longitude of the site
        :rtype: Tuple(float,float)
        """
        try:
            # site.get_property("location").to_latlon()
            return (
                self.get_topology_site(site_name).get_property("location").to_latlon()
            )
        except Exception as e:
            # logging.warning(f"Failed to get location postal {site_name}")
            return 0, 0

    def get_location_postal(self, site_name: str) -> str:
        """
        Gets the location of a site by postal address

        :param site_name: site name
        :type site_name: String
        :return: postal address of the site
        :rtype: String
        """
        try:
            return self.get_topology_site(site_name).location.postal
        except Exception as e:
            # logging.debug(f"Failed to get location postal {site_name}")
            return ""

    def get_host_capacity(self, site_name: str) -> int:
        """
        Gets the number of worker hosts at the site

        :param site_name: site name
        :type site_name: String
        :return: host count
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacities.unit
        except Exception as e:
            # logging.debug(f"Failed to get host count {site_name}")
            return 0

    def get_cpu_capacity(self, site_name: str) -> int:
        """
        Gets the total number of cpus at the site

        :param site_name: site name
        :type site_name: String
        :return: cpu count
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacities.cpu
        except Exception as e:
            # logging.debug(f"Failed to get cpu capacity {site_name}")
            return 0

    def get_core_capacity(self, site_name: str) -> int:
        """
        Gets the total number of cores at the site

        :param site_name: site name
        :type site_name: String
        :return: core count
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacities.core
        except Exception as e:
            # logging.debug(f"Failed to get core capacity {site_name}")
            return 0

    def get_core_allocated(self, site_name: str) -> int:
        """
        Gets the number of currently allocated cores at the site

        :param site_name: site name
        :type site_name: String
        :return: core count
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacity_allocations.core
        except Exception as e:
            # logging.debug(f"Failed to get cores allocated {site_name}")
            return 0

    def get_core_available(self, site_name: str) -> int:
        """
        Gets the number of currently available cores at the site

        :param site_name: site name
        :type site_name: String
        :return: core count
        :rtype: int
        """
        try:
            return self.get_core_capacity(site_name) - self.get_core_allocated(
                site_name
            )
        except Exception as e:
            # logging.debug(f"Failed to get cores available {site_name}")
            return self.get_core_capacity(site_name)

    def get_ram_capacity(self, site_name: str) -> int:
        """
        Gets the total amount of memory at the site in GB

        :param site_name: site name
        :type site_name: String
        :return: ram in GB
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacities.ram
        except Exception as e:
            # logging.debug(f"Failed to get ram capacity {site_name}")
            return 0

    def get_ram_allocated(self, site_name: str) -> int:
        """
        Gets the amount of memory currently  allocated the site in GB

        :param site_name: site name
        :type site_name: String
        :return: ram in GB
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacity_allocations.ram
        except Exception as e:
            # logging.debug(f"Failed to get ram allocated {site_name}")
            return 0

    def get_ram_available(self, site_name: str) -> int:
        """
        Gets the amount of memory currently  available the site in GB

        :param site_name: site name
        :type site_name: String
        :return: ram in GB
        :rtype: int
        """
        try:
            return self.get_ram_capacity(site_name) - self.get_ram_allocated(site_name)
        except Exception as e:
            # logging.debug(f"Failed to get ram available {site_name}")
            return self.get_ram_capacity(site_name)

    def get_disk_capacity(self, site_name: str) -> int:
        """
        Gets the total amount of disk available the site in GB

        :param site_name: site name
        :type site_name: String
        :return: disk in GB
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacities.disk
        except Exception as e:
            # logging.debug(f"Failed to get disk capacity {site_name}")
            return 0

    def get_disk_allocated(self, site_name: str) -> int:
        """
        Gets the amount of disk allocated the site in GB

        :param site_name: site name
        :type site_name: String
        :return: disk in GB
        :rtype: int
        """
        try:
            return self.get_topology_site(site_name).capacity_allocations.disk
        except Exception as e:
            # logging.debug(f"Failed to get disk allocated {site_name}")
            return 0

    def get_disk_available(self, site_name: str) -> int:
        """
        Gets the amount of disk available the site in GB

        :param site_name: site name
        :type site_name: String
        :return: disk in GB
        :rtype: int
        """
        try:
            return self.get_disk_capacity(site_name) - self.get_disk_allocated(
                site_name
            )
        except Exception as e:
            # logging.debug(f"Failed to get disk available {site_name}")
            return self.get_disk_capacity(site_name)

    def get_fablib_manager(self):
        return self.fablib_manager

    def update(self, force_refresh: bool = False):
        """
        Update the available resources by querying the FABRIC services

        """
        logging.info(f"Updating available resources")
        return_status, topology = (
            self.get_fablib_manager().get_slice_manager().resources(force_refresh=force_refresh)
        )
        if return_status != Status.OK:
            raise Exception(
                "Failed to get advertised_topology: {}, {}".format(
                    return_status, topology
                )
            )

        self.topology = topology

    def get_topology(self, update: bool = False) -> AdvertisedTopology:
        """
        Not intended for API use
        """
        if update or self.topology is None:
            self.update()

        return self.topology

    def get_site_list(self, update: bool = False) -> List[str]:
        """
        Gets a list of all sites by name

        :param update: (optional) set to True update available resources
        :type update: bool
        :return: list of site names
        :rtype: List[String]
        """
        if update or self.topology is None:
            self.update()

        rtn_sites = []
        for site_name, site in self.topology.sites.items():
            rtn_sites.append(site_name)

        return rtn_sites

    def get_link_list(self, update: bool = False) -> List[str]:
        """
        Gets a list of all links by name

        :param update: (optional) set to True update available resources
        :type update: bool
        :return: list of link names
        :rtype: List[String]
        """
        if update:
            self.update()

        rtn_links = []
        for link_name, link in self.topology.links.items():
            rtn_links.append(link_name)

        return rtn_links

    def site_to_json(self, site):
        return json.dumps(self.site_to_dict(site), indent=4)

    def site_to_dict(self, site):
        site_name = site.name
        return {
            "name": site.name,
            "state": self.get_state(site_name),
            "address": self.get_location_postal(site_name),
            "location": self.get_location_lat_long(site_name),
            "hosts": self.get_host_capacity(site_name),
            "cpus": self.get_cpu_capacity(site_name),
            "cores_available": self.get_core_available(site_name),
            "cores_capacity": self.get_core_capacity(site_name),
            "cores_allocated": self.get_core_capacity(site_name)
            - self.get_core_available(site_name),
            "ram_available": self.get_ram_available(site_name),
            "ram_capacity": self.get_ram_capacity(site_name),
            "ram_allocated": self.get_ram_capacity(site_name)
            - self.get_ram_available(site_name),
            "disk_available": self.get_disk_available(site_name),
            "disk_capacity": self.get_disk_capacity(site_name),
            "disk_allocated": self.get_disk_capacity(site_name)
            - self.get_disk_available(site_name),
            "nic_basic_available": self.get_component_available(
                site_name, "SharedNIC-ConnectX-6"
            ),
            "nic_basic_capacity": self.get_component_capacity(
                site_name, "SharedNIC-ConnectX-6"
            ),
            "nic_basic_allocated": self.get_component_capacity(
                site_name, "SharedNIC-ConnectX-6"
            )
            - self.get_component_available(site_name, "SharedNIC-ConnectX-6"),
            "nic_connectx_6_available": self.get_component_available(
                site_name, "SmartNIC-ConnectX-6"
            ),
            "nic_connectx_6_capacity": self.get_component_capacity(
                site_name, "SmartNIC-ConnectX-6"
            ),
            "nic_connectx_6_allocated": self.get_component_capacity(
                site_name, "SmartNIC-ConnectX-6"
            )
            - self.get_component_available(site_name, "SmartNIC-ConnectX-6"),
            "nic_connectx_5_available": self.get_component_available(
                site_name, "SmartNIC-ConnectX-5"
            ),
            "nic_connectx_5_capacity": self.get_component_capacity(
                site_name, "SmartNIC-ConnectX-5"
            ),
            "nic_connectx_5_allocated": self.get_component_capacity(
                site_name, "SmartNIC-ConnectX-5"
            )
            - self.get_component_available(site_name, "SmartNIC-ConnectX-5"),
            "nvme_available": self.get_component_available(site_name, "NVME-P4510"),
            "nvme_capacity": self.get_component_capacity(site_name, "NVME-P4510"),
            "nvme_allocated": self.get_component_capacity(site_name, "NVME-P4510")
            - self.get_component_available(site_name, "NVME-P4510"),
            "tesla_t4_available": self.get_component_available(
                site_name, "GPU-Tesla T4"
            ),
            "tesla_t4_capacity": self.get_component_capacity(site_name, "GPU-Tesla T4"),
            "tesla_t4_allocated": self.get_component_capacity(site_name, "GPU-Tesla T4")
            - self.get_component_available(site_name, "GPU-Tesla T4"),
            "rtx6000_available": self.get_component_available(site_name, "GPU-RTX6000"),
            "rtx6000_capacity": self.get_component_capacity(site_name, "GPU-RTX6000"),
            "rtx6000_allocated": self.get_component_capacity(site_name, "GPU-RTX6000")
            - self.get_component_available(site_name, "GPU-RTX6000"),
            "a30_available": self.get_component_available(site_name, "GPU-A30"),
            "a30_capacity": self.get_component_capacity(site_name, "GPU-A30"),
            "a30_allocated": self.get_component_capacity(site_name, "GPU-A30")
            - self.get_component_available(site_name, "GPU-A30"),
            "a40_available": self.get_component_available(site_name, "GPU-A40"),
            "a40_capacity": self.get_component_capacity(site_name, "GPU-A40"),
            "a40_allocated": self.get_component_capacity(site_name, "GPU-A40")
            - self.get_component_available(site_name, "GPU-A40"),
            "fpga_u280_available": self.get_component_available(site_name, "FPGA-Xilinx-U280"),
            "fpga_u280_capacity": self.get_component_capacity(site_name, "FPGA-Xilinx-U280"),
            "fpga_u280_allocated": self.get_component_capacity(site_name, "FPGA-Xilinx-U280")
                             - self.get_component_available(site_name, "FPGA-Xilinx-U280"),
        }

    def site_to_dictXXX(self, site):
        site_name = site.name
        return {
            "name": {"pretty_name": "Name", "value": site.name},
            "address": {
                "pretty_name": "Address",
                "value": self.get_location_postal(site_name),
            },
            "location": {
                "pretty_name": "Location",
                "value": self.get_location_lat_long(site_name),
            },
            "hosts": {
                "pretty_name": "Hosts",
                "value": self.get_host_capacity(site_name),
            },
            "cpus": {"pretty_name": "CPUs", "value": self.get_cpu_capacity(site_name)},
            "cores_available": {
                "pretty_name": "Cores Available",
                "value": self.get_core_available(site_name),
            },
            "cores_capacity": {
                "pretty_name": "Cores Capacity",
                "value": self.get_core_capacity(site_name),
            },
            "cores_allocated": {
                "pretty_name": "Cores Allocated",
                "value": self.get_core_capacity(site_name)
                - self.get_core_available(site_name),
            },
            "ram_available": {
                "pretty_name": "RAM Available",
                "value": self.get_ram_available(site_name),
            },
            "ram_capacity": {
                "pretty_name": "RAM Capacity",
                "value": self.get_ram_capacity(site_name),
            },
            "ram_allocated": {
                "pretty_name": "RAM Allocated",
                "value": self.get_ram_capacity(site_name)
                - self.get_ram_available(site_name),
            },
            "disk_available": {
                "pretty_name": "Disk Available",
                "value": self.get_disk_available(site_name),
            },
            "disk_capacity": {
                "pretty_name": "Disk Capacity",
                "value": self.get_disk_capacity(site_name),
            },
            "disk_allocated": {
                "pretty_name": "Disk Allocated",
                "value": self.get_disk_capacity(site_name)
                - self.get_disk_available(site_name),
            },
            "nic_basic_available": {
                "pretty_name": "Basic NIC Available",
                "value": self.get_component_available(
                    site_name, "SharedNIC-ConnectX-6"
                ),
            },
            "nic_basic_capacity": {
                "pretty_name": "Basic NIC Capacity",
                "value": self.get_component_capacity(site_name, "SharedNIC-ConnectX-6"),
            },
            "nic_basic_allocated": {
                "pretty_name": "Basic NIC Allocated",
                "value": self.get_component_capacity(site_name, "SharedNIC-ConnectX-6")
                - self.get_component_available(site_name, "SharedNIC-ConnectX-6"),
            },
            "nic_connectx_6_available": {
                "pretty_name": "ConnectX-6 Available",
                "value": self.get_component_available(site_name, "SmartNIC-ConnectX-6"),
            },
            "nic_connectx_6_capacity": {
                "pretty_name": "ConnectX-6 Capacity",
                "value": self.get_component_capacity(site_name, "SmartNIC-ConnectX-6"),
            },
            "nic_connectx_6_allocated": {
                "pretty_name": "ConnectX-6 Allocated",
                "value": self.get_component_capacity(site_name, "SmartNIC-ConnectX-6")
                - self.get_component_available(site_name, "SmartNIC-ConnectX-6"),
            },
            "nic_connectx_5_available": {
                "pretty_name": "ConnectX-5 Available",
                "value": self.get_component_available(site_name, "SmartNIC-ConnectX-5"),
            },
            "nic_connectx_5_capacity": {
                "pretty_name": "ConnectX-5 Capacity",
                "value": self.get_component_capacity(site_name, "SmartNIC-ConnectX-5"),
            },
            "nic_connectx_5_allocated": {
                "pretty_name": "ConnectX-5 Allocated",
                "value": self.get_component_capacity(site_name, "SmartNIC-ConnectX-5")
                - self.get_component_available(site_name, "SmartNIC-ConnectX-5"),
            },
            "nvme_available": {
                "pretty_name": "NVMe Available",
                "value": self.get_component_available(site_name, "NVME-P4510"),
            },
            "nvme_capacity": {
                "pretty_name": "NVMe Capacity",
                "value": self.get_component_capacity(site_name, "NVME-P4510"),
            },
            "nvme_allocated": {
                "pretty_name": "NVMe Allocated",
                "value": self.get_component_capacity(site_name, "NVME-P4510")
                - self.get_component_available(site_name, "NVME-P4510"),
            },
            "tesla_t4_available": {
                "pretty_name": "Tesla T4 Available",
                "value": self.get_component_available(site_name, "GPU-Tesla T4"),
            },
            "tesla_t4_capacity": {
                "pretty_name": "Tesla T4 Capacity",
                "value": self.get_component_capacity(site_name, "GPU-Tesla T4"),
            },
            "tesla_t4_allocated": {
                "pretty_name": "Tesla T4 Allocated",
                "value": self.get_component_capacity(site_name, "GPU-Tesla T4")
                - self.get_component_available(site_name, "GPU-Tesla T4"),
            },
            "rtx6000_available": {
                "pretty_name": "RTX6000 Available",
                "value": self.get_component_available(site_name, "GPU-RTX6000"),
            },
            "rtx6000_capacity": {
                "pretty_name": "RTX6000 Capacity",
                "value": self.get_component_capacity(site_name, "GPU-RTX6000"),
            },
            "rtx6000_allocated": {
                "pretty_name": "RTX6000 Allocated",
                "value": self.get_component_capacity(site_name, "GPU-RTX6000")
                - self.get_component_available(site_name, "GPU-RTX6000"),
            },
            "a30_available": {
                "pretty_name": "A30 Available",
                "value": self.get_component_available(site_name, "GPU-A30"),
            },
            "a30_capacity": {
                "pretty_name": "A30 Capacity",
                "value": self.get_component_capacity(site_name, "GPU-A30"),
            },
            "a30_allocated": {
                "pretty_name": "A30 Allocated",
                "value": self.get_component_capacity(site_name, "GPU-A30")
                - self.get_component_available(site_name, "GPU-A30"),
            },
            "a40_available": {
                "pretty_name": "A40 Available",
                "value": self.get_component_available(site_name, "GPU-A40"),
            },
            "a40_capacity": {
                "pretty_name": "A40 Capacity",
                "value": self.get_component_capacity(site_name, "GPU-A40"),
            },
            "a40_allocated": {
                "pretty_name": "A40 Allocated",
                "value": self.get_component_capacity(site_name, "GPU-A40")
                - self.get_component_available(site_name, "GPU-A40"),
            },
            "fpga_u280_available": {
                "pretty_name": "FPGA U280 Available",
                "value": self.get_component_available(site_name, "FPGA-Xilinx-U280"),
            },
            "fpga_u280_capacity": {
                "pretty_name": "FPGA U280 Capacity",
                "value": self.get_component_capacity(site_name, "FPGA-Xilinx-U280"),
            },
            "fpga_u280_allocated": {
                "pretty_name": "FPGA U280 Allocated",
                "value": self.get_component_capacity(site_name, "FPGA-Xilinx-U280")
                         - self.get_component_available(site_name, "FPGA-Xilinx-U280"),
            },
        }

    def list_sites(
        self,
        output=None,
        fields=None,
        quiet=False,
        filter_function=None,
        pretty_names=True,
    ):
        table = []
        for site_name, site in self.topology.sites.items():
            table.append(self.site_to_dict(site))

        if pretty_names:
            pretty_names_dict = self.site_pretty_names
        else:
            pretty_names_dict = {}

        return self.get_fablib_manager().list_table(
            table,
            fields=fields,
            title="Sites",
            output=output,
            quiet=quiet,
            filter_function=filter_function,
            pretty_names_dict=pretty_names_dict,
        )


class Links(Resources):
    link_pretty_names = {
        "site_names": "Sites",
        "node_id": "Link Name",
        "link_capacity_Gbps": "Capacity (Gbps)",
        "link_layer": "Link Layer",
    }

    def __init__(self, fablib_manager):
        """
        Constructor
        :return:
        """
        super().__init__(fablib_manager)

    def __str__(self) -> str:
        """
        Creates a tabulated string of all the links.

        Intended for printing available resources.

        :return: Tabulated string of available resources
        :rtype: String
        """
        table = []
        for _, link in self.topology.links.items():
            iface = link.interface_list[0]
            site_names = iface.name.split("_")
            if iface.type.name == "TrunkPort" and "HundredGig" not in site_names[0]:
                table.append(
                    [
                        tuple(site_names),
                        link.node_id,
                        iface.capacities.bw if iface.capacities else "N/A",
                        link.layer,
                    ]
                )

        return tabulate(
            table,
            headers=[
                "site_names",
                "node_id",
                "link_capacity_Gbps",
                "link_layer",
            ],
        )

    def link_to_dict(self, link: link.Link, iface: interface.Interface) -> dict:
        """
        Converts the link resources to a dictionary.

        Intended for printing links in table format.

        :return: collection of link properties
        :rtype: dict
        """
        return {
            "site_names": tuple(iface.name.split("_")),
            "node_id": link.node_id,
            "link_capacity_Gbps": iface.capacities.bw if iface.capacities else "N/A",
            "link_layer": link.layer,
        }

    def list_links(
        self,
        output=None,
        fields=None,
        quiet=False,
        filter_function=None,
        pretty_names=True,
    ) -> object:
        """
        Print a table of link resources in pretty format.

        :return: formatted table of resources
        :rtype: object
        """
        table = []
        for _, link in self.topology.links.items():
            iface = link.interface_list[0]
            site_names = iface.name.split("_")
            if iface.type.name == "TrunkPort" and "HundredGig" not in site_names[0]:
                table.append(self.link_to_dict(link, iface))

        if pretty_names:
            pretty_names_dict = self.link_pretty_names
        else:
            pretty_names_dict = {}

        return self.get_fablib_manager().list_table(
            table,
            fields=fields,
            title="Links",
            output=output,
            quiet=quiet,
            filter_function=filter_function,
            pretty_names_dict=pretty_names_dict,
        )


class FacilityPorts(Resources):
    link_pretty_names = {
        "site_name": "Site",
        "node_id": "Link Name",
        "vlan_range": "VLAN Range",
        "link_layer": "Link Layer",
    }

    def __init__(self, fablib_manager):
        """
        Constructor
        :return:
        """
        super().__init__(fablib_manager)

    def __str__(self) -> str:
        """
        Creates a tabulated string of all the links.

        Intended for printing available resources.

        :return: Tabulated string of available resources
        :rtype: String
        """
        table = []
        for _, link in self.topology.links.items():
            iface = link.interface_list[0]
            site_names = iface.name.split("_")
            if iface.type.name == "FacilityPort":
                table.append(
                    [
                        tuple(site_names),
                        link.node_id,
                        iface.labels.vlan_range,
                        link.layer,
                    ]
                )

        return tabulate(
            table,
            headers=[
                "site_name",
                "node_id",
                "vlan_range",
                "link_layer",
            ],
        )

    def fp_to_dict(self, link: link.Link, iface: interface.Interface) -> dict:
        """
        Converts the link resources to a dictionary.

        Intended for printing links in table format.

        :return: collection of link properties
        :rtype: dict
        """
        return {
            "site_name": tuple(iface.name.split("_")),
            "node_id": link.node_id,
            "vlan_range": iface.labels.vlan_range if iface.labels else "N/A",
            "link_layer": link.layer,
        }

    def list_facility_ports(
        self,
        output=None,
        fields=None,
        quiet=False,
        filter_function=None,
        pretty_names=True,
    ) -> object:
        """
        Print a table of link resources in pretty format.

        :return: formatted table of resources
        :rtype: object
        """
        table = []
        for _, link in self.topology.links.items():
            iface = link.interface_list[0]
            if iface.type.name == "FacilityPort":
                table.append(self.fp_to_dict(link, iface))

        if pretty_names:
            pretty_names_dict = self.link_pretty_names
        else:
            pretty_names_dict = {}

        return self.get_fablib_manager().list_table(
            table,
            fields=fields,
            title="Facility Ports",
            output=output,
            quiet=quiet,
            filter_function=filter_function,
            pretty_names_dict=pretty_names_dict,
        )
